/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: usb_core.h
 * Description: Core USB constants and data structures definitions.
 * Developer: Dimitar Dimitrov ( dinuxbg __on__ gmail.com )
 *
 * Last change: $Date: 2008-01-03 17:43:40 +0200 (четвъртък, 03 Януари 2008) $
 * Revision: $Revision: 105 $
 * Id: $Id: usb_core.h 105 2008-01-03 15:43:40Z dimitar $
 *
 ****************************************************************************/

#ifndef USB_CORE_H
#define USB_CORE_H

#ifndef __AVR__
  #error "This module requires AVR target processor! Anything else probably won't work (endianness, bitfield positions, alignement issues)"
#endif /* AVR */

/** USB recipient codes. */
#define USB_RECIPIENT_DEVICE			0x00
#define	USB_RECIPIENT_INTERFACE			0x01
#define USB_RECIPIENT_ENDPOINT			0x02
#define USB_RECIPIENT_OTHER			0x03

/** USB request type codes. */
#define USB_REQUEST_TYPE_STANDARD		0x00
#define USB_REQUEST_TYPE_CLASS			0x01
#define USB_REQEUST_TYPE_VENDOR			0x02

/** Standard request codes. */
#define USB_STDREQ_GET_STATUS			0x00
#define USB_STDREQ_CLEAR_FEATURE		0x01
#define USB_STDREQ_RESERVED_1			0x02
#define USB_STDREQ_SET_FEATURE			0x03
#define USB_STDREQ_RESERVED_2			0x04
#define USB_STDREQ_SET_ADDRESS			0x05
#define USB_STDREQ_GET_DESCRIPTOR		0x06
#define USB_STDREQ_SET_DESCRIPTOR		0x07
#define USB_STDREQ_GET_CONFIGURATION		0x08
#define USB_STDREQ_SET_CONFIGURATION		0x09
#define USB_STDREQ_GET_INTERFACE		0x0A
#define USB_STDREQ_SET_INTERFACE		0x0B
#define USB_STDREQ_SYNCH_FRAME			0x0C

/** Descriptor types. */
#define USB_DESCRTYPE_DEVICE			0x01
#define USB_DESCRTYPE_CONFIGURATION		0x02
#define USB_DESCRTYPE_STRING			0x03
#define USB_DESCRTYPE_INTERFACE			0x04
#define USB_DESCRTYPE_ENDPOINT			0x05
#define USB_DESCRTYPE_DEVICE_QUALIFIER		0x06
#define USB_DESCRTYPE_OTHER_SPEED_CONFIGURATION	0x07
#define USB_DESCRTYPE_INTERFACE_POWER		0x08
#define USB_DESCRTYPE_OTG			0x09
#define USB_DESCRTYPE_DEBUG			0x0A
#define USB_DESCRTYPE_INTERFACE_ASSOCIATION	0x0B

/** Standard Feature Selectors. */
#define USB_STDFEAT_DEVICE_REMOTE_WAKEUP	1
#define USB_STDFEAT_ENDPOINT_HALT		0
#define USB_STDFEAT_TEST_MODE			2

/** Endpoint type attribute bits definitions. */
#define USB_EPTYPE_CONTROL			0x00
#define USB_EPTYPE_ISOCHRONOUS			0x01
#define USB_EPTYPE_BULK				0x02
#define USB_EPTYPE_INTERRUPT			0x03

/** Endpoint direction bit constant for IN. */
#define USB_EPDIR_IN				0x1
/** Endpoint direction bit constant for OUT. */
#define USB_EPDIR_OUT				0x0
/** Endpoint direction bit constant for control endpoints. */
#define USB_EPDIR_CONTROL			USB_EPDIR_OUT

/** Control endpoint number. */
#define USB_EPN_CONTROL				0x00

#define USB_REL_20				0x0200
#define USB_REL_11				0x0110

/** USB device classes (vendor specific) */
#define USB_DEVCLASS_NONE			0x00
#define USB_DEVCLASS_IAD_CAPABLE		0xEF
#define USB_DEVCLASS_VENDOR			0xFFU
#define USB_DEVSUBCLASS_NONE			0x00
#define USB_DEVSUBCLASS_VENDOR			0x00
#define USB_DEVSUBCLASS_IAD_CAPABLE		0x02
#define USB_DEVPROT_NONE			0x00
#define USB_DEVPROT_IAD_CAPABLE			0x01
#define USB_INTRFCLASS_VENDOR			0xFFU
#define USB_INTRFSUBCLASS_VENDOR		0x00
#define USB_INTRFPROT_NONE			0x00

/** Convert current expressed in miliampers to USB current units. */
#define USB_POWER_mA_TO_USBUNITS(CURRENT)	(CURRENT/2)

/** 
 * USB SETUP DATA packet structure.
 *
 * @note AVR is little endian, so is USB.
 */
struct usb_setup_packet {
	struct {
		uint8_t recipient : 5;	/**< Recipient. */
		uint8_t type : 2;	/**< Request type. */
		uint8_t dir_d2h : 1;	/**< Data transfer direction
	       				 *	0 = Host-to-device
					 *	1 = Device-to-host
					 */
	} request_type;			/**< Characteristics of request. */
	uint8_t request;		/**< Specific request. */
	uint16_t value;			/**< Generic 16-bit value. */
	uint16_t index;			/**< Generic 16-bit value. */
	uint16_t length;		/**< Number of bytes to transfer if
					 * there is a Data stage. */
};

/**
 * Device descriptor structure.
 */
struct usb_device_descriptor {
	uint8_t length;			/**< Size of this descriptor in bytes*/
	uint8_t descriptor_type;	/**< DEVICE descriptor type. */
	uint16_t usbrel;		/**< USB specification release number
					 * in BCD format. */
	uint8_t device_class;		/**< Class code. Normally zero so that
					 * each interface within a 
					 * configuration specifies its own
					 * class information. */
	uint8_t device_sub_class;	/**< Subclass code. */
	uint8_t device_protocol;	/**< Protocol code. */
	uint8_t max_packet_size0;	/**< Maximum packet size for endpoint
					 * zero (only 8,16,32 or 64 are valid)*/
	uint16_t id_vendor;		/**< Vendro ID. */
	uint16_t id_product;		/**< Product ID. */
	uint16_t device_rel;		/**< Device release number in BCD. */
	uint8_t i_manifacturer;		/**< Index of string descriptor
					 * describing manifacturer. */
	uint8_t i_product;		/**< Index of string descriptor
					 * describing product. */
	uint8_t i_serial_number;	/**< Index of string descriptor
					 * describing device's serial number */
	uint8_t num_configurations;	/**< Number of possible 
					 * configurations. */
};

/**
 * Configuration descriptor structure.
 */
struct usb_configuration_descriptor {
	uint8_t length;			/**< Size of this descriptor in bytes*/
	uint8_t descriptor_type;	/**< CONFIGURATION descriptor type. */
	uint16_t total_length;		/**< Total length of data returned for 
					 * this configuration. Includes the 
					 * combined length of all descriptors
					 * (cfg,intrf,ep,class/vendor specific)
					 * returned for this configuration. */
	uint8_t num_interfaces;		/**< Number of interfaces supported by
					 * this configuration. */
	uint8_t configuration_value;	/**< Value to use as an argument to the
					 * SetConfiguration request to select
					 * this configuration. */
	uint8_t i_configuration;	/**< Index of string descriptor 
					 * describing this configuration. */
	struct {
		uint8_t reserved_0 : 5;		/**< Reset to zero. */
		uint8_t remote_wakeup : 1;
		uint8_t self_powered : 1;
		uint8_t reserved_1 : 1;		/**< Set to one. */
	} attributes;			/**< Configuration characteristics. */
	uint8_t max_power;		/**< Maximum power consumption of 
					 * the USB device from the bus. */
};

/**
 * Interface descriptor structure.
 */
struct usb_interface_descriptor {
	uint8_t length;			/**< Size of this descriptor in bytes*/
	uint8_t descriptor_type;	/**< INTERFACE descriptor type. */
	uint8_t interface_number;	/**< Number of this interface. Zero-
					 * based value identifying the index
					 * in the array of concurrent
					 * interfaces supported by this 
					 * configuration. */
	uint8_t alternate_setting;	/**< Value used to select this 
					 * alternate setting for the interface
					 * identified in the prior field.*/
	uint8_t num_endpoints;		/**< Number of endpoints used by this
					 * interface (excluding EP zero). */
	uint8_t interface_class;	/**< Class code. */
	uint8_t interface_sub_class;	/**< Sub class code. */
	uint8_t interface_protocol;	/**< Protocol code. */
	uint8_t i_interface;		/**< Index of string descriptor
					 * describing this interface. */
};


/**
 * Endpoint descriptor structure.
 */
struct usb_endpoint_descriptor {
	uint8_t length;			/**< Size of this descriptor in bytes*/
	uint8_t descriptor_type;	/**< ENDPOINT descriptor type. */
	struct {
		uint8_t epnum : 4;	/**< The endpoint number. */
		uint8_t reserved_0 : 3;	/**< Reset to zero. */
		uint8_t dir_isin : 1;	/**< Direction (ignored for control EP)
					 *	0 = OUT endpoint
					 *	1 = IN endpoint; */
	} endpoint_address;		/**< The address of the endpoint. */
	struct {
		uint8_t transfer_type : 2; 
		uint8_t synchronization_type :2;
		uint8_t usage_type : 2;
		uint8_t reserved_0 : 2;
	} attributes;			/**< Endpoint's attributes. */
	uint16_t max_packet_size;	/**< Maximum packet size for this EP.*/
	uint8_t interval;		/**< Interval for polling EP for data
					 * transfers. Expressed in (u)frames */
};

/**
 * Generic string descriptor structure. Please use the 
 * @a USB_CREATE_STRING_DESCRIPTOR macro to define string descriptors.
 */
struct usb_string_descriptor {
	uint8_t length;			/**< Size of this descriptor in bytes*/
	uint8_t descriptor_type;	/**< STRING descriptor type. */
	uint16_t string[1];		/**< Placeholder for the unicode
					 * string array. */
};

/**
 * Creates a string descriptor variable named @a NAME with @a SIZE number
 * of unicode characters.
 *
 * @param NAME Descriptor uniqueue name.
 * @param SIZE Number of unicode characters in the string descriptor.
 * @param ... A list of unicode characters which will be stored in the 
 * descriptor. Normally user will give a list of <i>char</i> constants that 
 * will be promoted to <i/>uint16_t</i>.
 *
 * @example USB_CREATE_STRING_DESCRIPTOR(manif, 6, 'o','l','i','m','e','x');
 */
#define USB_CREATE_STRING_DESCRIPTOR(NAME, SIZE, ... ) \
	struct usb_strdescr_ ## NAME { \
		uint8_t length; \
		uint8_t descriptor_type; \
		uint16_t wchar_array[(SIZE)]; \
	} NAME PROGMEM = { \
		sizeof(struct usb_strdescr_ ## NAME), \
		USB_DESCRTYPE_STRING, \
		{ __VA_ARGS__ } \
	}

/**
 * Standard Interface Association descriptor structure. 
 */
struct usb_interface_association_descriptor {
	uint8_t length;			/**< Size of this descriptor in bytes*/
	uint8_t descriptor_type;	/**< IAD descriptor type. */
	uint8_t first_interface;	/**< First associated interface. */
	uint8_t interface_count;	/**< Number of _contiguois_ interfaces
					 * that are associated with this
					 * function. */
	uint8_t function_class;		/**< Class code, assigned by USB-IF,
					 * zero value not allowed. */
	uint8_t function_subclass;	/**< Subclass code. */
	uint8_t function_protocol;	/**< Protocol code,assigned by USB-IF*/
	uint8_t i_function;		/**< Index of string descriptor,
					 * describing this function. */
};

/** User has to instantiate the device descriptor. */
extern struct usb_device_descriptor usb_device_descriptor PROGMEM;

/** User has to instantiate the string descriptors table. */
extern struct usb_string_descriptor *(usb_strings_table[]) PROGMEM;

/** User has to provide function that returns the number of entries in the
 * string descriptor table. This is needed to prevent reading beyond the
 * end of the table. */
extern uint8_t usb_get_strtbl_size(void);

/** User has to specify a structure, containing the configuration data that 
 * will be returned upon GetConfiguration Descriptor standard request. */
struct usb_configuration_block;

/** User has to instantiate the configuration block structure. */
extern struct usb_configuration_block usb_configuration_block PROGMEM;

/** Handle SETUP control request. For internal use only. */
extern void usb_core_setupreq(void);

/** Points to the currently selected configuration. */
extern uint8_t usb_current_configuration;

#endif /* USB_CORE_H */

