/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: usb_hw.c
 * Description: Hardware-dependant USB functionality.
 * Developer: Dimitar Dimitrov ( dinuxbg  __on__ gmail.com )
 *
 * Last change: $Date: 2008-01-02 16:29:10 +0200 (сряда, 02 Януари 2008) $
 * Revision: $Revision: 103 $
 * Id: $Id: usb_hw.c 103 2008-01-02 14:29:10Z dimitar $
 *
 ****************************************************************************/

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/sleep.h>


#include "usb_conf.h"
#include "usb_hw.h"
#include "usb_core.h"
#include "usb.h"

static void usb_poweron(void);
static void usb_poweroff(void);
static void usb_suspend(void);

/**
 * Configure and reset the control endpoint. 
 * @return Zero for success, non-zero value on error.
 */
static int8_t usb_ctrlep_init(void)
{
	usb_ep_select(USB_EPN_CONTROL);

	usb_ep_reset(USB_EPN_CONTROL);
	usb_ep_reset_datatoggle();
	if(!usb_ep_enabled()) {
		return usb_ep_activate(
				USB_EPN_CONTROL,
				USB_EPTYPE_CONTROL,
				USB_EPDIR_CONTROL,
				USB_CTRLEP_MAX_PACKET_SIZE,
				0	/* no doublebank */
			);
	} else {
		return 0;
	}
}

/** 
 * Various flags set by the generic USB interrupt and processed by the 
 * usb polling function.
 */
static volatile struct {
	uint8_t connected : 1;	/**< Device is expected to be connected. */
	uint8_t reset : 1;	/**< USB reset has been detected. */
	uint8_t pwr_on_rst : 1;	/**< Whether USB reset has been received after
				 * power up. */
} usb_flags;


/** 
 * Poll for standard requests on EP0 and handle them. For non-standard
 * requests call the user-defined function usbcb_setupreq().
 *
 * This function also processes generic USB events like reset, suspend and
 * wakeup by calling the appropriate user-defined functions.
 */
void usb_std_poll(void)
{
	if(usb_flags.connected) {
		if(!usb_check_bus_power()) {
			USB_ATOMIC_OPER((usb_flags.connected = 0, 
					 usb_flags.pwr_on_rst = 0));
			usb_poweroff();
			usb_current_configuration = 0;
		}
	} else {
		usb_current_configuration = 0;
		if(usb_check_bus_power()) {
			USB_ATOMIC_OPER(usb_flags.connected = 1);
			usb_poweron();
		}
	}

	if(usb_flags.reset) {
		USB_ATOMIC_OPER((usb_flags.reset = 0,
				 usb_flags.pwr_on_rst = 1));
		usb_ctrlep_init();
		usb_current_configuration = 0;
	}

	usb_ep_select(USB_EPN_CONTROL);
	if(usb_setupreq_ready() && usb_flags.pwr_on_rst) {
		usb_core_setupreq();
	}
}

/**
 * USB Generic Interrupt Vector.
 * Used to catch asynchronous events like reset, suspend, wakeup.
 */
ISR(USB_GEN_vect)
{
	const uint8_t intstat = usb_genintr_fetch();
	usb_genintr_clear(intstat);

	if(usb_intst_sof(intstat)) {
		usbcb_sof();
	}

	if(usb_intst_suspend(intstat)) {
		usb_suspend();
	}

	if(usb_intst_reset(intstat)) {
		usb_flags.reset = 1;
	}
}

/*
ISR(USB_COM_vect)
{
}
*/

/** Handle USB power on. This function actually starts the USB macro cell 
 * (the device controller) along with the needed PLL initialization fluff.
 * At the end the USB device controller is attached.
 */
static void usb_poweron(void)
{
	/* Configure and enable PLL */
	usb_pll_stop();
	usb_pll_start();
	/* Check PLL lock */
	while(!usb_pll_locked());
	/* Enable USB interface */
	usb_macro_enable();
	usb_resume_clock();
	/* Configure USB interface (USB Endpoint 0 configuratin) */
	usb_ctrlep_init();
	/* Attach USB device */
	usb_attach();
	usb_inten_sof();
	usb_inten_reset();
	usb_inten_suspend();
}

/** Handle USB power off, that is detaching the device from the host. */
static void usb_poweroff(void)
{
	/* Detach USB device */
	usb_detach();
	/* Disable USB interface */
	usb_macro_disable();
	/* Disable PLL */
	usb_pll_stop();
}

/** 
 * Handle USB suspend. This function shuts down the USB PLL clock and enables
 * the USB wakeup interrupt. 
 */
static void usb_suspend(void)
{
	/* we will be woken up by a USB reset */
	usb_current_configuration = 0;
	/* warning - why don't we enable the following line?
	 * usb_flags.pwr_on_rst = 0; */

	/* If we have power then there is no need for suspend. */
	if(usb_is_self_powered()) return;

	/* Clear Suspend Bit - done by the USB GEN ISR */

	/* Be sure to have interrupts enabled (WAKEUPE) to exit sleep mode */
	usb_inten_wakeup();
	usb_intclr_wakeup();
	usb_intdis_suspend();

	/* Set USB suspend clock */
	usb_freeze_clock();

	/* Disable PLL */
	usb_pll_stop();

	usbcb_suspend();

	/* Put the MCU in sleep mode */
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	sleep_enable();
	sei();	/* we do need that in order to wakeup */
	sleep_cpu();
	cli();
	sleep_disable();


	usbcb_wakeup();

	/* Enable PLL */
	usb_pll_start();
	/* Wait PLL lock 
	 * @warning We wait a lot of time here! */
	while(!usb_pll_locked());
	/* Clear USB suspend clock */
	usb_resume_clock();
	/* Disable further interrupts on USB pad change. */
	usb_intclr_wakeup();
	usb_intdis_wakeup();
	/* Be prepared for other suspends. */
	usb_inten_suspend();
}

/** 
 * Activates the currently active endpoint and allocates memory for its 
 * buffers. Please use the more user-friendly macro usb_ep_activate instead. 
 */
int8_t usb_ep_activate_conf(uint8_t config0x, uint8_t config1x)
{
	/* Configure:
	 * 	- endpoint direction;
	 * 	- endpoint type;
	 * 	- Not-Yet-Disable feature
	 */
	UECFG0X = config0x;

	/* Configure:
	 * 	- endpoint size
	 * 	- bank parametrization
	 */
	UECFG1X = (UECFG1X & (1<<ALLOC)) | config1x;

	usb_epmem_allocate();

	if(UESTA0X & (1<<CFGOK)) {
		/* endpoint activated */
		return 0;
	} else {
		/* error */
		return 1;
	}
}

