/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: usbscsi.h
 * Description: USB Mass Storage SCSI Module
 * Developer: Dimitar Dimitrov ( dinuxbg@gmail.com )
 *
 * Last change: $Date: 2007-12-05 17:37:13 +0200 (сряда, 05 Декември 2007) $
 * Revision: $Revision: 73 $
 * Id: $Id: usbscsi.h 73 2007-12-05 15:37:13Z dimitar $
 *
 ****************************************************************************/

#ifndef USBSCSI_H
#define USBSCSI_H

#include "massbulk.h"

/** SBC commands, required by the USB Mass Storage standard. */
enum sbc_opcode {
	SBC_OP_TEST_UNIT_READY 			= 0x00,
	SBC_OP_REQUEST_SENSE 			= 0x03,
	SBC_OP_FORMAT_UNIT 			= 0x04,
	SBC_OP_INQUIRY 				= 0x12,
	SBC_OP_MODE_SENSE_6			= 0x1A,
	SBC_OP_PREVENT_ALLOW_MEDIUM_REMOVAL	= 0x1E,
	SBC_OP_READ_CAPACITY 			= 0x25,
	SBC_OP_READ_10 				= 0x28,
	SBC_OP_WRITE_10 			= 0x2A,
	SBC_OP_VERIFY 				= 0x2F,
	SBC_OP_MODE_SENSE_10 			= 0x5A
};

/** Periphery device type codes. */
enum sbc_periph_dev_type {
	SBC_DEVTYPE_SCSI_DIRECT_ACCESS = 0x00,
	SBC_DEVTYPE_CDROM = 0x05,
	SBC_DEVTYPE_OPTICAL = 0x07,
	SBC_DEVTYPE_RBC_DIRECT_ACCESS = 0x0E
};

/** SBC INQUIRY command block. */
struct sbc_inquiry_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */

	uint8_t reserved : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */

	uint8_t reserved_1;
	uint8_t reserved_2;		/**< Allocation length MSB byte??? */
	uint8_t allocation_length;	/**< Must be sizeof(sbc_inquiry_data)*/
	uint8_t reserved_3;
};

/** SBC PREVENT_ALLOW_MEDIUM_REMOVAL command block. */
struct sbc_prevent_allow_medium_removal_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */

	uint8_t reserved : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */

	uint8_t reserved_1;
	uint8_t reserved_2;

	uint8_t prevent : 2;
	uint8_t : 6;
	uint8_t control;
};

/** Version of the SBC implementation. */
enum sbc_version {
	SBC_VER_NO_STANDARD_CONFORMANCE 	= 0x00,
	SBC_VER_SPC				= 0x03,
	SBC_VER_SPC2				= 0x04,
	SBC_VER_SPC3				= 0x05,
};

/** SBC INQUIRY answer block. */
struct sbc_inquiry_data {
	enum sbc_periph_dev_type devtype;/**< Periphery Device Type */

	uint8_t reserved_0 : 7;
	uint8_t removable : 1;		/**< Whether medium is removable */

	enum sbc_version version;	/**< Should be set to 0 */

	uint8_t response_data_format : 4;/**< Should be 2. */
	uint8_t hisup : 1;
	uint8_t normaca : 1;
	uint8_t obsolete_1 : 1;
	uint8_t obsolete_2 : 1;

	uint8_t additional_length;	/**< Equal to sizeof(sbc_inquiry_data)
					 * minus 4. */
	uint8_t sccs : 1;
	uint8_t acc : 1;
	uint8_t tpgs : 2;
	uint8_t pc3 : 1;
	uint8_t reserved_1 : 2;
	uint8_t protect : 1;

	uint8_t bque : 1;
	uint8_t encserv : 1;
	uint8_t vs : 1;
	uint8_t multip : 1;
	uint8_t mchngr : 1;
	uint8_t obsolete_0 : 2;
	uint8_t addr16 : 1;

	uint8_t reserved_2;

	uint8_t vendor_id[8];		/**< ASCII Vendor Identification. */
	uint8_t product_id[16];		/**< ASCII Product Identification. */
	uint8_t product_rev[4];		/**< ASCII Product Revision. */
};

/** SBC Read Command Block. */
struct sbc_read_10_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_0 : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t lba_b3;			/**< Logical Block Address .*/
	uint8_t lba_b2;
	uint8_t lba_b1;
	uint8_t lba_b0;
	uint8_t reserved_1;
	uint8_t transfer_length_b1;
	uint8_t transfer_length_b0;
	uint8_t reserved_2;
};

/** SBC Request Sense Command Block. */
struct sbc_request_sense_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_0 : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t reserved_1;
	uint8_t reserved_2;
	uint8_t allocation_length;	/**< sizeof(sbc_request_sense_data) */
	uint8_t control;
};

/** Request Sense Error Code. */
enum sbc_reqsense_errcode {
	SBC_REQSENSE_CURRENT_ERROR = 0x70,
	SBC_REQSENZE_DEFERRED_ERROR = 0x71
};

/** Request Sense Sense Key Codes. */
enum sbc_sense_key {
	/** Nothing to be reported. */
	SBC_SENSEKEY_NO_SENSE = 0x00,
	/**< Command completed, but error condition had been recovered. */
	SBC_SENSEKEY_RECOVERED_ERROR = 0x01,
	/** Logical Unit is not accessible. */
	SBC_SENSEKEY_NOT_READY = 0x02,
	/** Medium fault. */
	SBC_SENSEKEY_MEDIUM_ERROR = 0x03,
	/** Non-recoverable hardware failure. */
	SBC_SENSEKEY_HARDWARE_ERROR = 0x04,
	/** Indicates that:
	 * 	a) LUN is incorrect;
	 * 	b) Command had an invalid task attribute;
	 * 	c) LUN is unable to process the command;
	 * 	d) Illegal parameter in CDB;
	 * 	e) Illegal parameter in the additional parameters. 
	 */
	SBC_SENSEKEY_ILLEGAL_REQUEST = 0x05,
	SBC_SENSEKEY_UNIT_ATTENTION = 0x06,
	/** Read/Write attempted on a protected block. */
	SBC_SENSEKEY_DATA_PROTECT = 0x07,
	SBC_SENSEKEY_BLANK_CHECK = 0x08,
	SBC_SENSEKEY_VENDOR_SPECIFIC = 0x09,
	SBC_SENSEKEY_COPY_ABORTED = 0x0A,
	SBC_SENSEKEY_ABORTED_COMMAND = 0x0B,
	SBC_SENSEKEY_VOLUME_OVERFLOW = 0x0D,
	SBC_SENSEKEY_MISCOMPARE = 0x0E,
};

/** 
 * Pairs of additional sense key/qualifier codes. MSB byte is the sense key,
 * and the LSB byte is the qualifier.
 */
enum sbc_additional_sense_pair {
	SBC_ADDSP_NO_ADDITIONAL_SENSE_INFORMATION = 0x0000,
	SBC_ADDSP_PERIPHERAL_DEVICE_WRITE_FAULT = 0x0300,
	SBC_ADDSP_UNRECOVERED_READ_ERROR = 0x1100,
	SBC_ADDSP_MISCOMPARE_DURING_VERIFY_OPERATION = 0x1D00,
	SBC_ADDSP_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE = 0x2100,
	SBC_ADDSP_INVALID_FIELD_IN_COMMAND_PACKET = 0x2400,
	SBC_ADDSP_WRITE_PROTECTED = 0x2700,
	SBC_ADDSP_MEDIUM_NOT_PRESENT = 0x3A00
};

/** Answer format for the Request Sense Operation. */
struct sbc_request_sense_data {
	uint8_t errcode : 7;		/** Use sbc_reqsense_errcode values.*/
	uint8_t valid : 1;
	uint8_t segment_number;
	uint8_t sense_key : 4;
	uint8_t reserved_1 : 1;
	uint8_t ili : 1;		/**< Incorrect Length Indicator */
	uint8_t reserved_2 : 2;
	uint8_t information_b3;
	uint8_t information_b2;
	uint8_t information_b1;
	uint8_t information_b0;
	uint8_t additional_sense_length;	/* n-7 */
	uint8_t cmd_specific_info_b3;
	uint8_t cmd_specific_info_b2;
	uint8_t cmd_specific_info_b1;
	uint8_t cmd_specific_info_b0;
	uint8_t additional_sense_code;
	uint8_t additional_sense_code_qualifier;	/* optional */
	uint8_t reserved_3[4];
};

/** Test Unit Ready Control Block. */
struct sbc_test_unit_ready_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_0 : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t reserved_1;
	uint8_t reserved_2;
	uint8_t reserved_3;
	uint8_t reserved_4;
};

/** Mode Sense 6 Control Block. */
struct sbc_mode_sense_6_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_1 : 3;
	uint8_t dbd : 1;
	uint8_t reserved_2 : 1;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t page_code;
	uint8_t subpage_code;
	uint8_t plist_length;
	uint8_t control;
};

/** Mode Sense 10 Control Block. */
struct sbc_mode_sense_10_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_1 : 3;
	uint8_t dbd : 1;
	uint8_t reserved_2 : 1;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t page_code;
	uint8_t reserved_3;
	uint8_t reserved_4;
	uint8_t reserved_5;
	uint8_t reserved_6;
	uint8_t plist_length_b1;
	uint8_t plist_length_b0;
	uint8_t reserved_7;
};

/** Mode Sense Parameter Header. */
struct sbc_mode_param_header {
	uint8_t mode_data_length_b1;	/* should be 0x00 */
	uint8_t mode_data_length_b0;	/* should be 0x26 */
	uint8_t reserved_1;
	uint8_t reserved_2;
	uint8_t reserved_3;
	uint8_t reserved_4;
	uint8_t block_descr_length_b1;
	uint8_t block_descr_length_b0;	/* 0 or 8, should be 0! */
};

/** Mode Sense Page code. */
enum sbc_page_code {
	SBC_PAGECODE_FLEXIBLE_DISK = 0x05,
	SBC_PAGECODE_RETURN_ALL = 0x3F
};

/** Mode Sense Page. */
struct sbc_mode_page {
	enum sbc_page_code page_code;	/**< Should be flexible disk page */
	uint8_t page_length;		/**< Value in bytes */
	uint8_t unused_1;
	uint8_t unused_2;
	uint8_t num_heads;
	uint8_t sec_per_track;
	uint8_t bytes_per_sector_b1;
	uint8_t bytes_per_sector_b0;
	uint8_t num_cyl_b1;
	uint8_t num_cyl_b0;
	uint8_t unused_3[22];
};

/**
 * Mode Sense Paramter list. That's the information returned on MODE SENSE
 * request. 
 */
struct sbc_mode_param_list {
	struct sbc_mode_param_header header;
	/*struct sbc_block_desriptor block_descr;*/
	struct sbc_mode_page page;	/* we support only one! */
};

/** SCSI Read Capacity Command Block. */
struct sbc_read_capacity_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_0 : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t reserved_1;
	uint8_t reserved_2;
	uint8_t reserved_3;
	uint8_t reserved_4;
	uint8_t reserved_5;
	uint8_t reserved_6;
	uint8_t reserved_7;
	uint8_t reserved_8;
};

/** SCSI Read Capacity Answer Block. */
struct sbc_read_capacity_data {
	uint8_t last_lba_b3;
	uint8_t last_lba_b2;
	uint8_t last_lba_b1;
	uint8_t last_lba_b0;
	uint8_t block_length_b3;
	uint8_t block_length_b2;
	uint8_t block_length_b1;
	uint8_t block_length_b0;
};

/** Low-level format control block request. */
struct sbc_format_unit_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t defect_list_format : 3; /* should be 0x07 */
	uint8_t cmplist : 1;		/* should be 0 */
	uint8_t fmtdata : 1;		/* should be 1 */
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t vendor_specific;
	uint8_t interleave_b1;
	uint8_t interleave_b0;
	uint8_t reserved;
};

/** Verify control block. */
struct sbc_verify_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_1 : 1;
	uint8_t bytechk : 1;
	uint8_t reserved_2 : 3;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t lba_b3;
	uint8_t lba_b2;
	uint8_t lba_b1;
	uint8_t lba_b0;
	uint8_t reserved_3;
	uint8_t verif_length_b1;
	uint8_t verif_length_b0;
	uint8_t reserved_4;
};

/** Write 10 control block. */
struct sbc_write_10_cb {
	enum sbc_opcode opcode;		/**< Operation Code. */
	uint8_t reserved_0 : 5;
	uint8_t lun : 3;		/**< Logical Unit Number. */
	uint8_t lba_b3;
	uint8_t lba_b2;
	uint8_t lba_b1;
	uint8_t lba_b0;
	uint8_t reserved_1;
	uint8_t transfer_length_b1;
	uint8_t transfer_length_b0;
	uint8_t reserved_2;
};

extern enum ms_csw_status sbc_decode(const uint8_t cb[16], uint32_t *data_length, const uint8_t dir_d2h);

#endif /* USBSCSI_H */

